# Free FFT and convolution (Python)
# Copyright (c) 2012 Nayuki Minase
# http://nayuki.eigenstate.org/page/free-small-fft-in-multiple-languages
#
# (MIT License)
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# - The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
# - The Software is provided "as is", without warranty of any kind, express or
#   implied, including but not limited to the warranties of merchantability,
#   fitness for a particular purpose and noninfringement. In no event shall the
#   authors or copyright holders be liable for any claim, damages or other
#   liability, whether in an action of contract, tort or otherwise, arising from,
#   out of or in connection with the Software or the use or other dealings in the
#   Software.
#

import cmath


#
# Computes the discrete Fourier transform (DFT) of the given complex vector, returning the result as a new vector.
# Set 'inverse' to True if computing the inverse transform. This DFT does not perform scaling, so the inverse is not a true inverse.
# The vector can have any length. This is a wrapper function.
#
def transform(vector, inverse=False):
    n = len(vector)
    if n > 0 and n & (n - 1) == 0:  # Is power of 2
        return transform_radix2(vector, inverse)
    else:  # More complicated algorithm for aribtrary sizes
        return transform_bluestein(vector, inverse)


#
# Computes the discrete Fourier transform (DFT) of the given complex vector, returning the result as a new vector.
# The vector's length must be a power of 2. Uses the Cooley-Tukey decimation-in-time radix-2 algorithm.
#
def transform_radix2(vector, inverse):
    # Initialization
    n = len(vector)
    levels = _log2(n)
    exptable = [
        cmath.exp((2j if inverse else -2j) * cmath.pi * i / n) for i in range(n // 2)
    ]
    vector = [
        vector[_reverse(i, levels)] for i in range(n)
    ]  # Copy with bit-reversed permutation

    # Radix-2 decimation-in-time FFT
    size = 2
    while size <= n:
        halfsize = size // 2
        tablestep = n // size
        for i in range(0, n, size):
            k = 0
            for j in range(i, i + halfsize):
                temp = vector[j + halfsize] * exptable[k]
                vector[j + halfsize] = vector[j] - temp
                vector[j] += temp
                k += tablestep
        size *= 2
    return vector


#
# Computes the discrete Fourier transform (DFT) of the given complex vector, returning the result as a new vector.
# The vector can have any length. This requires the convolution function, which in turn requires the radix-2 FFT function.
# Uses Bluestein's chirp z-transform algorithm.
#
def transform_bluestein(vector, inverse):
    # Find a power-of-2 convolution length m such that m >= n * 2 + 1
    n = len(vector)
    m = 1
    while m < n * 2 + 1:
        m *= 2

    exptable = [
        cmath.exp((1j if inverse else -1j) * cmath.pi * (i * i % (n * 2)) / n)
        for i in range(n)
    ]  # Trigonometric table
    a = [x * y for (x, y) in zip(vector, exptable)] + [0] * (
        m - n
    )  # Temporary vectors and preprocessing
    b = [
        (exptable[min(i, m - i)].conjugate() if (i < n or m - i < n) else 0)
        for i in range(m)
    ]
    c = convolve(a, b, False)[:n]  # Convolution
    for i in range(n):  # Postprocessing
        c[i] *= exptable[i]
    return c


#
# Computes the circular convolution of the given real or complex vectors, returning the result as a new vector. Each vector's length must be the same.
# realoutput=True: Extract the real part of the convolution, so that the output is a list of floats. This is useful if both inputs are real.
# realoutput=False: The output is always a list of complex numbers (even if both inputs are real).
#
def convolve(x, y, realoutput=True):
    assert len(x) == len(y)
    n = len(x)
    x = transform(x)
    y = transform(y)
    for i in range(n):
        x[i] *= y[i]
    x = transform(x, inverse=True)

    # Scaling (because this FFT implementation omits it) and postprocessing
    if realoutput:
        for i in range(n):
            x[i] = x[i].real / n
    else:
        for i in range(n):
            x[i] /= n
    return x


# Returns the integer whose value is the reverse of the lowest 'bits' bits of the integer 'x'.
def _reverse(x, bits):
    y = 0
    for i in range(bits):
        y = (y << 1) | (x & 1)
        x >>= 1
    return y


# Returns the integer y such that 2^y == x, or raises an exception if x is not a power of 2.
def _log2(x):
    i = 0
    while True:
        if 1 << i == x:
            return i
        elif 1 << i > x:
            raise ValueError("Not a power of 2")
        else:
            i += 1



from pyglet.media.synthesis import SynthesisSource, WhiteNoise
import ctypes
import random
from reiz._audio.fft import transform


class BrownNoise(SynthesisSource):
    """A waveform. of brown noise, i.e. with 1/f² 
    :Parameters:
        `duration` : float
            The length, in seconds, of audio that you wish to generate.
        `sample_rate` : int
            Audio samples per second. (CD quality is 44100).
    """

    def __init__(self, duration, **kwargs):
        super().__init__(duration, **kwargs)
        
    def _generate_data(self, num_bytes):
        samples = num_bytes >> 1
        amplitude = 32767
        fxx  = []
        for i in range(samples):
            tmp = random.normalvariate(0, 1) + 1j * random.normalvariate(0, 1) 
            fxx.append(tmp / (i+1))
        data = (ctypes.c_short * samples)()
        envelope = self._envelope_generator
        signal = transform(fxx, inverse=True)
        for i in range(samples):
            data[i] = int(signal[i].real  * amplitude * next(envelope))
        return data


import reiz


def test_visual_library(canvas):
    "contains Image, Cross, Mural and Trapezoid"
    for item in reiz.visual.library.__dict__.values():
        item.draw(canvas)


def test_colors(canvas):
    "also draws a background"
    from reiz._visual.colors import COLORS
    from reiz._visual.colors import resolve_rgb

    for c in COLORS:
        resolve_rgb("white", c)
        reiz.visual.Background(color=c).draw(canvas)


def test_smoke_complex(canvas):
    reiz._visual.complex.Line().draw(canvas)
    reiz._visual.complex.Polygon().draw(canvas)
    reiz._visual.complex.Bar().draw(canvas)
    reiz._visual.complex.Circle().draw(canvas)
    reiz._visual.complex.Cylinder().draw(canvas)
    reiz._visual.complex.Mural().draw(canvas)
    reiz._visual.complex.Trapezoid().draw(canvas)
    reiz._visual.complex.Cross().draw(canvas)


def test_circle(canvas):
    v = reiz._visual.complex.Circle(zoom=0.1, color="white", stroke=1, opacity=0.1)
    assert v.color == [1, 1, 1, 0.1]
    assert v.stroke == 1
    v.draw(canvas)


def test_mural(canvas):
    m = reiz._visual.complex.Mural(text="test")
    m.set_color("white")
    assert m.color == [1, 1, 1, 1]
    assert repr(m) == "Mural('test')"
    m.draw(canvas)


def test_cross(canvas):
    m = reiz._visual.complex.Cross(zoom=1, color=(1, 0, 0))
    assert repr(m) == "Cross(zoom=1, color=(1, 0, 0))"


def test_iteration(canvas):
    for v in reiz._visual.complex.Line():
        v.draw(canvas)

    for v in [reiz._visual.complex.Line(), reiz._visual.complex.Line()]:
        v.draw(canvas)


def test_read_folder(canvas):
    lib = reiz.visual.read_folder()
    assert lib.__dict__.get("logo", None) is not None



"""Creating and managing visual stimuli
"""
import os
from os import environ as _env
from reiz._visual.complex import Mural, Image, Cross, Circle, Background, Line, Bar
from reiz._visual.complex import Polygon, Trapezoid, Cylinder
from reiz._visual.colors import COLORS, get_color
from types import SimpleNamespace
from typing import Dict, NewType, Any
from pathlib import Path
from pkg_resources import resource_filename

libConf = NewType(
    "libConf", Dict[str, Dict[str, Any]]
)  #: Dict[str, Dict[str, Any], a dictionary of types and respective keywords arguments

_defaults = libConf(
    {
        "Mural": {
            "post": {"text": "Run endet"},
            "pre": {"text": "Run beginnt"},
            "eo": {"text": "Augen offen"},
            "ec": {"text": "Augen zu"},
            "ready": {"text": "Bereit machen"},
            "los": {"text": "Los"},
            "go": {"text": "Go!"},
            "imagine": {"text": "Bewegung vorstellen"},
            "move": {"text": "Bewegung starten"},
            "count": {"text": "Zahlen berechnen"},
            "relax": {"text": "Entspannen"},
            "rating": {
                "text": "0 - 1 - 2 - 3 - 4 - 5 - 6 - 7 - 8 - 9 - 10",
                "fontsize": 0.75,
            },
        },
        "Cross": {"fixation": {"color": "white"},},
        "Image": {"logo": {"imgpath": resource_filename(__name__, "data/logo.png")},},
    }
)  #: libConf


def read_folder(folder: str = None) -> SimpleNamespace:
    """"read image files in a the specified folder into a library

    args
    ----
    folder:str
        the path to the folder with the images


    The function selects all files ending in .png and .jpg. It collects them  into a SimpleNamespace, so you can use dot notation to select the images. These fields are named based on the filename. Some basic sanitization is performed, but is nonetheless best if the files already have names which could be called safely from python. For example, files should be named 'image_file.png' instead of 'image-file.png'.
          
    """

    folder = folder or resource_filename(__name__, "data/")

    def sanitize(fname: str) -> str:
        return fname.replace("-", "_").replace(" ", "_").replace(":", "_")

    conf = dict()
    conf["Image"] = dict()
    pth = Path(folder).expanduser().absolute()
    print(pth)
    for file in pth.glob("*.png"):
        print(file)
        conf["Image"][sanitize(file.stem)] = {"imgpath": str(file)}
    for file in pth.glob("*.jpg"):
        print(file)
        conf["Image"][sanitize(file.stem)] = {"imgpath": str(file)}
    return make_library(conf)


def make_library(settings: libConf = None, failraise=False) -> SimpleNamespace:
    """create a library of visual stimuli from a dictionary of arguments

    args
    ----
    settings:
        a dictionary of types with kwargs appropriate to the respective type
        as dictionary
    failraise:bool
        raise an exception if a key is not recognized. defaults to False

    returns
    -------

    library:
        a library of instances of stimuli which can be adressed in dot.notation
    """

    if settings is None:
        settings = _defaults
    library = dict()
    for key in settings.keys():
        for name, args in settings[key].items():
            if key.lower() == "mural":
                library[name] = Mural(**args)
            elif key.lower() == "cross":
                library[name] = Cross(**args)
            elif key.lower() == "image":
                library[name] = Image(**args)
            else:
                if failraise:
                    raise (f"{key} with {args} can't be processed")

    library = SimpleNamespace(**library)
    return library


if not "DOC" in _env.keys():
    library = make_library()
else:  # pragma: no cover
    mock = dict()
    for t, item in _defaults.items():
        for k, v in item.items():
            mock[k] = t
    library = SimpleNamespace(**mock)  #: a library of visual stimuli
    print("Generating sphinx documentation. Skipping visual library")


__doc__ += f"""
During import, a basic library will be created. By default, it contains
the following set of visual stimuli: {', '.join([str(k) for k in library.__dict__])}. These stimuli can be addressed by dot.notation from :data:`reiz.visual.library`, for example like :data:`reiz.visual.library.go`

.. automodule:: reiz._visual.complex
   :members: Mural, Image, Cross, Circle, Background, Line, Bar, Polygon, Trapezoid, Cylinder

.. automodule:: reiz._visual.colors
   :members: COLORS, get_color

Create libraries of visual stimuli
..................................



"""
